//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//    Copyright (C) 2007, 2008 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
//
//==============================================================================
program lossyWAV;
//==============================================================================
{$APPTYPE CONSOLE}
{$DEFINE CPU386}
{$DEFINE MSWINDOWS}
{$R *.res}
{$N+ $A8}
//==============================================================================

uses
  Sysutils,
  Windows,
  Math,
  nparameter in 'units\nparameter.pas',
  nfft in 'units\nfft.pas',
  ntypes in 'units\ntypes.pas',
  nwav in 'units\nwav.pas',
  nrand in 'units\nrand.pas';

//==============================================================================
type
//==============================================================================
  fft_results_rec = packed record
    btr      : Integer;
    start    : integer;
    spread   : Integer;
    nminbin  : Integer;
    analysis : Integer;
    sminbin  : Double;
    savebin  : Double;
  end;

//==============================================================================
const
//==============================================================================
  precalc_analyses      =  6;
  spread_zones          =  4;
  Quality_Presets       = 10;
  skewing_amplitude     = 36;
  codec_block_size      = 512;

  //============================================================================
  // pre-calculated reference_threshold surface for no dither bit-reduction
  //============================================================================
  reference_threshold : array[1..precalc_analyses,1..32] of Integer =
  ((6,11,16,22,28,34,40,46,52,58,64,70,76,82,88,94,100,106,112,118,124,130,136,142,148,154,160,166,172,178,184,190),
   ( 9,14,19,25,31,37,43,49,55,61,67,73,79,85,91,97,103,109,115,121,127,133,139,145,152,158,164,170,176,182,188,194),
   (12,17,22,28,34,40,46,52,58,64,70,76,82,88,94,100,106,112,118,124,130,136,142,148,155,161,167,173,179,185,191,197),
   (15,20,25,31,37,43,49,55,61,67,73,79,85,91,97,103,109,115,121,127,133,139,145,152,158,164,170,176,182,188,194,200),
   (18,23,28,34,40,46,52,58,64,70,76,82,88,94,100,106,112,118,124,130,136,142,148,155,161,167,173,179,185,191,197,203),
   (21,26,31,37,43,49,55,61,67,73,79,85,91,97,103,109,115,121,127,133,139,145,152,158,164,170,176,182,188,194,200,206));

  //============================================================================
  // constants directly related to quality_integer settings.
  //============================================================================

  spreading_function_string         : string[precalc_analyses*(spread_zones+2)-1]='22222-22223-22224-12234-12245-12356';

  fft_analysis_string               : array[0..5] of string[precalc_analyses] = ('000000','000001','010001','011001','011101','011111');

  fft_bit_length                    : array[1..precalc_analyses] of Integer = (5,6,7,8,9,10);

  fft_length                        : array[1..precalc_analyses] of Integer = (32,64,128,256,512,1024);

  fft_length_recip                  : array[1..precalc_analyses] of Double  = (1/32,1/64,1/128,1/256,1/512,1/1024);

  quality_minimum_bits_to_keep      : array[0..Quality_Presets] of double = (2.9,2.95,3,3.125,3.25,3.375,3.5,3.625,3.75,3.875,4.0);

  quality_noise_threshold_shifts    : array[0..Quality_Presets] of Double  = (20,16,9,6,3,0,-2.4,-4.8,-7.2,-9.6,-12);

  quality_signal_to_noise_ratios    : array[0..Quality_Presets] of Double  = (18,22,23.5,23.5,23.5,25,28,31,34,37,40);

  quality_auto_fft32_on             : array[0..Quality_presets] of Integer = (0,0,0,1,1,1,1,1,1,1,1);

  quality_clips_per_channel         : array[0..Quality_Presets] of Integer = (3,3,3,3,2,1,0,0,0,0,0);

  //============================================================================
  // Frequency range over which to check for lowest amplitude signal
  //============================================================================
  Frequency_Limits                  : Array[0..spread_zones+1] of Double = (20,1378.125,3445.3125,8268.75,12403.125,16000);

  max_samples : array[0..bitstocalculate-1] of Integer = ($7FFFFFFF,$7FFFFFFE,$7FFFFFFC,$7FFFFFF8,$7FFFFFF0,$7FFFFFE0,$7FFFFFC0,$7FFFFF80,
                                                          $7FFFFF00,$7FFFFE00,$7FFFFC00,$7FFFF800,$7FFFF000,$7FFFE000,$7FFFC000,$7FFF8000,
                                                          $7FFF0000,$7FFE0000,$7FFC0000,$7FF80000,$7FF00000,$7FE00000,$7FC00000,$7F800000,
                                                          $7F000000,$7E000000,$7C000000,$78000000,$70000000,$60000000,$40000000,$00000000);

//==============================================================================
var
  WF_Arr                            : TwoPowerDoubleRecord;
  SK_Arr                            : TwoPowerDoubleRecord;
  SA_Arr                            : TwoPowerIntegerRecord;

  skewing_function   : array[1..precalc_analyses] of PtrDoubleVector  = (@SK_Arr.p5,@SK_Arr.p6,@SK_Arr.p7,@SK_Arr.p8,@SK_Arr.p9,@SK_Arr.pA);
  window_function    : array[1..precalc_analyses] of PtrDoubleVector  = (@WF_Arr.p5,@WF_Arr.p6,@WF_Arr.p7,@WF_Arr.p8,@WF_Arr.p9,@WF_Arr.pA);
  spreading_averages : array[1..precalc_analyses] of PtrIntegerVector = (@SA_Arr.p5,@SA_Arr.p6,@SA_Arr.p7,@SA_Arr.p8,@SA_Arr.p9,@SA_Arr.pA);

  STimer                            : STimer_Record;

  ex_log2x20                        : double = log2x20;
  ex_MaxDB                          : double = MaxDB;

  this_codec_block_channel_rms      : array[0..maxchannels-1] of double = (0,0,0,0,0,0,0,0);

  noise_threshold_shift             : Double;
  noise_threshold_shift_exp         : Double;
  scaling_factor                    : Double;
  scaling_factor_inv                : Double;
  snr_value                         : Double;
  snr_value_exp                     : Double;
  processing_rate                   : Double;
  this_minimum_bits_to_keep         : Double;

  fib_bins                          : Array[0..32] of integer = (0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0);

  this_max_sample                   : Integer;
  this_min_sample                   : Integer;

  empty_codec_blocks                : Integer = 0;
  quality_integer                   : Integer;
  detailled_output                  : Integer;
  samples_processed                 : Integer;

  //============================================================================

  bits_used_this_codec_block        : Double = 0;
  bits_used_this_codec_block_min    : Double = 0;
  bits_used_this_codec_block_avg    : Double = 0;
  bits_used_this_codec_block_channel  : Array[0..maxchannels-1] of Double;

  quality_double                    : Double = 0;
  quality_fraction                  : double;

  channel                           : Integer;
  channels                          : Integer;

  clipped_samples                   : Integer;
  current_codec_block               : Integer;
  codec_block_bytes                 : Integer;

  success                           : Integer;

  this_fft_result                   : fft_results_rec;

  btr_row_tot                       : Integer;

  analysis_number                   : Integer;
  last_print                        : Integer;
  half_mb_blocks                    : Integer;

  this_analysis_number              : Integer;

  this_spf                          : string[precalc_analyses*(spread_zones+2)-1];
  this_fft_string                   : string[7];

  bin_width                         : array[0..precalc_analyses] of Double;
  bin_time                          : array[0..precalc_analyses] of Double;

  fft_result                        : array[0..maxfftlength-1] of Double;

  analysis_blocks                   : array[1..precalc_analyses] of Integer;

  //============================================================================
  // For each FFT bit length store bits_to_remove with respect to calculated minimum dB of FFT result.
  threshold_index                   : array[1..precalc_analyses,0..255] of Integer;
  //============================================================================
  // Frequency bins corresponding to Frequency Limits for each analysis;
  Frequency_Bins                    : array[1..precalc_analyses,0..spread_zones+1] of Integer;
  lo_bins                           : array[1..precalc_analyses] of integer;
  hi_bins                           : array[1..precalc_analyses] of integer;
  max_bins                          : array[1..precalc_analyses] of integer;
  num_bins                          : array[1..precalc_analyses] of integer;
  // spreading_lengths are zero based; Array padded to speed-up assembly routines.
  spreading_lengths                 : array[1..precalc_analyses,0..spread_zones+1] of Integer;
  //============================================================================

  blocks_read                       : Integer;
  total_bits_removed                : Integer;
  total_bits_lost                   : Integer;

  //initial_setup;
  sa_lfb                            : Double;
  sa_mfb                            : Double;
  sa_dfb                            : Double;

  sa_i                              : Integer;
  sa_j                              : Integer;
  sa_k                              : Integer;
  last_filled                       : Integer;
  sa_bit                            : Integer;
  this_reference_threshold          : Double;
  total_overlap_length              : Integer;

  //set_divisor_string;
  sds_d                             : Double;
  sds_o                             : Integer;

  //process_this_codec_block;
  fft_underlap_length               : array[1..precalc_analyses] of Double;
  end_overlap_length                : array[1..precalc_analyses] of Integer;
  central_block_start               : array[1..precalc_analyses] of Integer;
  half_number_of_blocks             : array[1..precalc_analyses] of Integer;
  actual_analysis_blocks_start      : array[1..precalc_analyses] of Integer;

  actual_analysis_block_start       : Integer;
  this_fft_length                   : Integer;
  this_actual_analysis_blocks_start : Integer;
  this_fft_underlap_length          : Double;
  this_fft_length_recip             : Double;
  fft_average                       : Integer = 0;
  fft_minimum                       : Integer = 0;
  min_fft_result                    : fft_results_rec;

  //remove_bits
  Old_Btr                           : Integer;
  this_codec_block_clips            : Integer;
  maximum_clips_per_channel         : Integer;

  //spread
  fpu_control_word                  : Integer;
  num_bin                           : integer;

  //hexval
  ucaseval                          : integer;

  //fill_fft_array;
  this_Integer                      : Integer;
  btrd_integer                      : Integer;
  last_sample                       : double;

  saved_fft_results : array[0..maxchannels-1,1..precalc_analyses] of fft_results_rec = (((btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1)),
                                                                                        ((btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1)),
                                                                                        ((btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1)),
                                                                                        ((btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1)),
                                                                                        ((btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1)),
                                                                                        ((btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1)),
                                                                                        ((btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1)),
                                                                                        ((btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1),(btr:0;start:-1)));

//==============================================================================

  function Window_Value : Extended; Assembler;
  Asm
                fld     qword ptr powersoftwoin[1*real_size]
                fld1
                fild    sa_i
                fadd    st,st(2)
                fmul    this_fft_length_recip
                fldpi
                fmul
                fadd    st,st
                fcos
                fsub
                fmul
                fmul    scaling_factor
  End;

  //============================================================================
  // Check command line parameters from nparameter;
  //============================================================================
  procedure check_switches;
  begin
    QueryPerformanceFrequency(STimer.Frequency);
    QueryPerformanceCounter(STimer.Start);

    if quality<>-1 then
      quality_double:=quality
    else
      quality_double:=5;

    quality_integer:=floor(quality_double);

    if quality_integer=quality_double then
      parameter_string:='-q '+inttostr(quality_integer)
    else
      parameter_string:='-q '+floattostrf(quality_double,ffFixed,99,4);

    quality_fraction:=quality_double-quality_integer;

    codec_block_bytes:=codec_block_size * integer_size;

    if noisethreshold<>255 then
    begin
      noise_threshold_shift:=round(noisethreshold*1000)/1000;
      parameter_string:=parameter_string+' -nts '+floattostrf(noise_threshold_shift,ffFixed,99,3);
    end
    else
    begin
      if quality_integer=quality_presets then
        noise_threshold_shift:=quality_noise_threshold_shifts[quality_presets]
      else
        noise_threshold_shift:=quality_fraction*(quality_noise_threshold_shifts[quality_integer+1]-quality_noise_threshold_shifts[quality_integer])+quality_noise_threshold_shifts[quality_integer];
      noise_threshold_shift:=round(noise_threshold_shift*1000)/1000;
    end;
    noise_threshold_shift_exp:=power(2,noise_threshold_shift/Log2x20);

    if snr>-216 then
    begin
      snr_value:=round(snr*1000)/1000;
      parameter_string:=parameter_string+' -snr '+floattostrf(snr_value,ffFixed,99,3);
    end
    else
    begin
      if quality_integer=quality_presets then
        snr_value:=quality_signal_to_noise_ratios[quality_presets]
      else
        snr_value:=quality_fraction*(quality_signal_to_noise_ratios[quality_integer+1]-quality_signal_to_noise_ratios[quality_integer])+quality_signal_to_noise_ratios[quality_integer];
      snr_value:=round(snr_value*1000)/1000;
    end;
    snr_value_exp:=power(2,-snr_value/Log2x20);

    if minbits>-1 then
    begin
      this_minimum_bits_to_keep:=minbits;
      parameter_string:=parameter_string+' -minbits '+floattostrf(minbits,ffFixed,99,2);
    end
    else
      this_minimum_bits_to_keep:=quality_minimum_bits_to_keep[quality_integer];

    if scaling<>-1 then
    begin
      scaling_factor:=scaling;
      parameter_string:=parameter_string+' -scale '+floattostrf(scaling_factor,fffixed,99,6);
    end
    else
      scaling_factor:=1;

    if scaling_factor=0 then
      lossyWAVerror('Scale value must not be zero',$21);

    scaling_factor_inv:=1/scaling_factor;

    if fft_string<>'' then
    begin
      this_fft_string:=fft_string;
      parameter_string:=parameter_string+' -fft '+this_fft_string;
    end
    else
    begin
      if analyses=0 then
        this_analysis_number:=2
      else
        this_analysis_number:=analyses;

      this_fft_string:=fft_analysis_string[this_analysis_number];
      if this_analysis_number<>2 then
        parameter_string:=parameter_string+' -analyses '+inttostr(this_analysis_number);
    end;

    if (fft32=1) or (quality_auto_fft32_on[quality_integer]=1) then
    begin
      this_fft_string[1]:='1';
      parameter_string:=parameter_string+' -impulse';
    end;

    if noclips=1 then
    begin
      maximum_clips_per_channel:=0;
      parameter_string:=parameter_string+' -noclips ';
    end
    else
      maximum_clips_per_channel:=quality_clips_per_channel[min(quality_presets,quality_integer+ord(quality_fraction<>0))];

    if spf_string='' then
      this_spf:=spreading_function_string
    else
    begin
      this_spf:=spf_string;
      parameter_string:=parameter_string+' -spf '+this_spf;
    end;

    if lowpass<>-1 then
    begin
      parameter_string:=parameter_string + ' -lowpass '+floattostrf(lowpass,ffFixed,99,0);
      frequency_limits[spread_zones+1]:=round(lowpass);
    end;

    parameter_string:=parameter_string+crlf+#26#0;

    detailled_output:=detailledoutput;
    if detailled_output=1 then
      lossyWAVwarning('Detailled output mode enabled');

    if priority=1 then
      setpriorityclass(getcurrentprocess,$00004000); // BELOW_NORMAL_PRIORITY_CLASS

    if priority=2 then
      setpriorityclass(getcurrentprocess,$00000040); // IDLE_PRIORITY_CLASS
  end;

  function hexval(hvchar : char) : integer;
  begin
    hexval:=0;
    ucaseval:=ord(upcase(hvchar));
    case ucaseval of
      49..57,65..70 : hexval:=ucaseval-(48+ord(ucaseval>57)*7);
    else
      lossyWAVerror('Invalid character in -spf input: "'+hvchar+'".',$21);
    end;
  end;

  //============================================================================
  // Setup most of the internal arrays of values to be used in the analyses;
  //============================================================================
  procedure initial_setup;
  begin
    if (verbosity=1) and (silent=0) then
    begin
      write('Processing : '+WAVFileName+crlf);
      write('Format     : ',samplerate/1000:0:2,'kHz; ',channelcount,' ch.; ',bitspersample,' bit.',crlf);
      write('Progress   : ');
    end
    else
      if (silent=0) then
        write(WAVFileName,';');

    // dummy setting to stop compiler messages.......
    WF_Arr.p0[0]:=0;
    SK_Arr.p0[0]:=0;
    SA_Arr.p0[0]:=0;

    channels:=channelcount;
    clipped_samples:=0;

    this_max_sample:=powersoftwom1[bitspersample-1];
    this_min_sample:=-powersoftwo[bitspersample-1];

    for sa_i:=0 to bitstocalculate-1 do
      max_samples[sa_i]:=max_samples[sa_i] and this_max_sample;

    // Setup some internal constants

    sa_lfb:=log10(frequency_limits[0]); // 20Hz lower limit for skewing;
    sa_mfb:=log10(frequency_limits[2]); // 3445.3125Hz upper limit for skewing;
    sa_dfb:=sa_mfb-sa_lfb;              // skewing range;

    for analysis_number := 1 to precalc_analyses do
    begin
      this_fft_length:=fft_length[analysis_number];
      this_fft_length_recip:=fft_length_recip[analysis_number];

      for sa_i:=0 to spread_zones do
        spreading_lengths[analysis_number,sa_i]:=hexval(this_spf[(analysis_number-1)*(spread_zones+2)+sa_i+1])-1;

      end_overlap_length[analysis_number]:=min(this_fft_length div 2,codec_block_size);
      actual_analysis_blocks_start[analysis_number]:=-end_overlap_length[analysis_number];
      total_overlap_length:=codec_block_size+end_overlap_length[analysis_number]*2-this_fft_length;
      fft_underlap_length[analysis_number]:=this_fft_length div 2;
      analysis_blocks[analysis_number]:=max(0,floor(total_overlap_length/fft_underlap_length[analysis_number]));
      fft_underlap_length[analysis_number]:=total_overlap_length/max(1,analysis_blocks[analysis_number]);

      // Calculate actual analysis_time values for fft_lengths
      bin_width[analysis_number]:=samplerate*this_fft_length_recip;
      bin_time[analysis_number]:=1/bin_width[analysis_number];

      // Generate window_function and skewing function for each fft_length
      for sa_i:=0 to fft_length[analysis_number]-1 do
        window_function[analysis_number]^[sa_i]:=Window_Value;

      // Calculate which FFT bin corresponds to the low frequency limit
      for sa_i:=0 to spread_zones+1 do
      begin
        frequency_bins[analysis_number,sa_i]:=max(1,round(Frequency_Limits[sa_i]*bin_time[analysis_number])-1);
        if frequency_bins[analysis_number,sa_i]>this_fft_length div 2 then
          lossyWAVerror('frequency too high : '+floattostrf(frequency_limits[sa_i]/1000,ffFixed,99,3)+'kHz',$21);
      end;

      for sa_j:=spread_zones downto 0 do
        for sa_i:=frequency_bins[analysis_number,sa_j] to frequency_bins[analysis_number,sa_j+1] do
          spreading_averages[analysis_number]^[sa_i]:=spreading_lengths[analysis_number,sa_j];

      lo_bins[analysis_number]:=frequency_bins[analysis_number,0];
      hi_bins[analysis_number]:=frequency_bins[analysis_number,spread_zones+1];
      max_bins[analysis_number]:=hi_bins[analysis_number]-spreading_averages[analysis_number]^[hi_bins[analysis_number]];
      num_bins[analysis_number]:=max_bins[analysis_number]-lo_bins[analysis_number]+1;

      for sa_i:=0 to frequency_bins[analysis_number,spread_zones+1] do
      begin
        skewing_function[analysis_number]^[sa_i]:=power(10,(power(sin(HalfPi*Min(sa_dfb,max(0,log10(max(1,sa_i*bin_width[analysis_number]))-sa_lfb))/sa_dfb),0.75)-1)*skewing_amplitude/20);
      end;

      // Calculate Threshold_Index values using Reference_Threshold Values.
      last_filled:=0;
      for sa_bit:=0 to bitstocalculate-1 do
        while(last_filled<reference_threshold[analysis_number,sa_bit+1]) do
        begin
          threshold_index[analysis_number,last_filled]:=sa_bit;
          inc(last_filled);
        end;

    end; // calculating for each analysis_number

    half_mb_blocks:=524288 div channels div bytespersample div codec_block_size;
    last_print:=0;
    blocks_read:=1;
    total_bits_removed:=0;
  end;

  procedure Get_Elapsed_Time;
  Begin
    QueryPerformanceCounter(STimer.Stop);
    STimer.Elapsed:=(Stimer.Stop-STimer.Start)/STimer.Frequency;
    processing_rate:=samples_processed/samplerate/STimer.Elapsed;
  End;

  procedure write_cleanup;
  begin
    get_elapsed_time;

    if (verbosity=1)  and (silent=0)then
    begin
      if (detailled_output=1) then
        write(crlf);
      write(#13,'                                                                               ',#13,'Average    :');
    end;

    write(' ',powersoftwoin[20]*samples_processed*channels*bytespersample:0:2,'MB; ');
    if (silent=0) then
    begin
      if (verbosity=1) then
      begin
        write(total_bits_removed/blocks_read:0:4,' bits; ');
        if (warnings=0) then
          write(total_bits_lost/blocks_read:0:4,' lost; ');
      end
      else
        write(total_bits_removed/blocks_read:0:4,'; ');
    end;

    write(processing_rate:0:2,'x; '); //Overall processing rate;
    write(stimer.elapsed:0:2,'s;');   //Overall processing time;

    if (verbosity=0)  and (silent=0) then
      if (warnings=0) then
      begin
        if clipped_samples>0 then
          write(' C:',clipped_samples,';'); //Total clipped samples in file;
        if total_bits_lost>0 then
          write(' L:',total_bits_lost,';'); //Total bits lost due to clipping;
      end;

    write(crlf);

    if total_bits_lost>0 then
      lossyWAVwarning(floattostrf(total_bits_lost/blocks_read,ffFixed,99,4)+' bits not removed due to clipping.');

    if clipped_samples>0 then
      lossyWAVwarning(inttostr(clipped_samples)+' sample(s) clipped to limiting amplitude.');
  end;

  procedure fill_FFT_Input;
  asm
                push    ebx
                push    ebp
                push    edi
                push    esi

                xor     edi,edi

                mov     esi,analysis_number
                mov     esi,dword ptr window_function[esi*pointer_size-pointer_size]
                mov     eax,channel
                mov     eax,dword ptr channel_offset[eax*integer_size]
                mov     ebx,actual_analysis_block_start
                add     ebx,this_fft_length
                sub     ebx,1
                mov     edx,codec_block_size

      @@00:     mov     ecx,ebx

                cmp     ecx,edx
                jl      @@01

                sub     ecx,edx
                mov     ebp,dword ptr rotating_blocks_ptr[3*pointer_size]
                add     ebp,eax
                fild    dword ptr [ebp+ecx*integer_size]
                jmp     @@04

      @@01:     cmp     ecx,0
                jl      @@02

                mov     ebp,dword ptr rotating_blocks_ptr[2*pointer_size]
                add     ebp,eax
                fild    dword ptr [ebp+ecx*integer_size]
                jmp     @@04

      @@02:     add     ecx,edx
                cmp     ecx,0
                jl      @@03

                mov     ebp,dword ptr rotating_blocks_ptr[1*pointer_size]
                add     ebp,eax
                fild    dword ptr [ebp+ecx*integer_size]
                jmp     @@04

      @@03:     add     ecx,edx
                mov     ebp,dword ptr rotating_blocks_ptr[0*pointer_size]
                add     ebp,eax
                fild    dword ptr [ebp+ecx*integer_size]

      @@04:     fmul    qword ptr [esi+edi]
                fstp    qword ptr FFT_Array[edi]

                add     edi,real_size
                sub     ebx,1
                cmp     ebx,actual_analysis_block_start
                jge     @@00

                pop     esi
                pop     edi
                pop     ebp
                pop     ebx
  end;

  function Spread_Complex : integer; assembler;
  asm
                push    ebx
                push    ebp
                push    edi
                push    esi

  //============================================================================
  // Fill FFT_Result array with zero values.
  //============================================================================
                mov     eax,this_fft_length
                shr     eax,1

                fldz
                jmp     @@01

      @@00:       fst     qword ptr fft_result[eax*real_size]
      @@01:       sub     eax,1
                  jge     @@00

                fstp    st
  //============================================================================

                mov     eax,analysis_number
                mov     ebx,dword ptr lo_bins[eax*integer_size-integer_size]
                mov     edi,ebx
                mov     edx,dword ptr hi_bins[eax*integer_size-integer_size]
                mov     ebp,dword ptr max_bins[eax*integer_size-integer_size]
                mov     ecx,dword ptr num_bins[eax*integer_size-integer_size]
                mov     num_bin,ecx
                mov     esi,dword ptr skewing_function[eax*integer_size-integer_size]

  //============================================================================
  // Calculate magnitude of FFT_Array values and skew.
  //============================================================================
                shl     ebx,1
                shl     edx,1

      @@02:     fld     qword ptr fft_array[ebx*real_size]
                fmul    st,st
                fld     qword ptr fft_array[ebx*real_size+real_size]
                fmul    st,st
                fadd
                fsqrt
                fmul    qword ptr [esi+ebx*(real_size/2)]
                fstp    qword ptr fft_result[ebx*(real_size/2)]

      @@05:     add     ebx,2
                cmp     ebx,edx
                jle     @@02

  //============================================================================
  // Calculate spread minimum and average results
  //============================================================================
                mov     eax,analysis_number
                fld     ex_log2x20                  //
                fldz                                //rolling_average
                fld     ex_maxdb                    //this_minimum

                mov     esi,dword ptr spreading_averages[eax*integer_size-integer_size]

      @@2:        mov     ecx,edi
                  mov     ebx,[esi+ecx*integer_size]
                  shl     ecx,real_shift
                  mov     edx,ebx

                  fldz

      @@3:          fadd    qword ptr fft_result[ecx+edx*real_size]
                    sub     edx,1
                    jge     @@3

      @@4:        fmul    qword ptr OneOverX[ebx*real_size+real_size]
                  fld     st
                  fadd    st,st(3)
                  fstp    st(3)

                  fld     st(1)

                  fcom
                  fnstsw  ax
                  test    ah,$01
                  jz      @@5

                  fxch    st(1)
                  mov     this_fft_result.nminbin,edi

      @@5:        fstp    st
                  fstp    st(1)

                  add     edi,1
                  cmp     edi,ebp
                  jle     @@2

                fmul    noise_threshold_shift_exp
                fst     this_fft_result.sminbin

                fld     st(1)
                fmul    snr_value_exp
                fidiv   num_bin
                fst     this_fft_result.savebin
                fstp    st(2)

                fcom
                fnstsw  ax
                test    ah,$01
                jz      @@6

                fxch    st(1)

      @@6:      fstp    st

                fld1
                fcom
                fnstsw  ax
                test    ah,$01
                jnz     @@7

                fxch    st(1)

      @@7:      fstp    st
                fyl2x

                //store rounding style
                fstcw   word ptr FPU_control_word
                mov     ax,word ptr FPU_control_word
                //change rounding style to truncate toward zero, full precision
                or      byte ptr FPU_control_word+1,0Fh
                fldcw   word ptr FPU_control_word
                //store spreading_result_temp
                fistp   this_fft_result.spread
                //restore rounding style
                mov     word ptr FPU_control_word,ax
                fldcw   word ptr FPU_control_word

                mov    eax,this_fft_result.spread
  //============================================================================

                pop     esi
                pop     edi
                pop     ebp
                pop     ebx
  end;

  procedure remove_bits_output;
  var
    rb_j : Integer;
  Begin
    if (verbosity=1) and (silent=0) then
    begin
      if current_codec_block=0 then
        write(crlf,'Block    Time   00 01 02 03 04 05 06 07 08 09 10 11 12 13 14 15 Tot.'+crlf
                  +'===================================================================='+crlf);
      if current_codec_block and 15=0 then
      begin
        if current_codec_block>0 then
          write(BTR_row_tot:4,crlf);
        write(current_codec_block:5,current_codec_block*codec_block_size/samplerate:8:2,'s.');
        BTR_row_tot:=0;
      end;
      write(min_fft_result.btr:3);
      BTR_row_tot:=BTR_row_tot+min_fft_result.BTR;
    end
    else
    begin
      if current_codec_block=0 then
        write(crlf);
      if current_codec_block and 15=0 then
      begin
        if current_codec_block>0 then
          write(BTR_row_tot,crlf);
        write(current_codec_block,',',current_codec_block*codec_block_size/samplerate:0:2,'s.,');
        BTR_row_tot:=0;
      end;
      write(min_fft_result.btr,',');
      BTR_row_tot:=BTR_row_tot+min_fft_result.BTR;
    end;

    current_codec_block:=current_codec_block+1;

    if this_codec_block_size<codec_block_size then
      if (silent=0) then
      begin
        if verbosity=1 then
        begin
         if blocks_read and 15<>0 then
           begin
            for rb_j:=1 to 16-(blocks_read and 15) do
              write('  -');
            write(btr_row_tot:4,crlf);
            write('====================================================================');
          end;
        end
        else
        begin
          if blocks_read and 15<>0 then
          begin
            for rb_j:=1 to 16-(blocks_read and 15) do
              write(',');
            write(btr_row_tot,crlf);
          end;
        end;
      end;
  End;

  procedure remove_bits; assembler;
  asm
                push    ebx
                push    ebp
                push    esi
                push    edi

                mov     eax,min_fft_result.btr
                mov     old_btr,eax

      @@08:     mov     eax,min_fft_result.btr
                cmp     eax,1
                jl      @@26

                fild    dword ptr powersoftwo[eax*int64_size]
                fld     qword ptr powersoftwoin[eax*real_size]
                fmul    scaling_factor
                fild    dword ptr max_samples[eax*integer_size]
                fild    dword ptr this_min_sample

                xor     eax,eax
                xor     ecx,ecx
                mov     this_codec_block_clips,eax

                mov     esi,channels

      @@09:       sub     esi,1
                  jl      @@25

                  mov     ebx,dword ptr channel_offset[esi*integer_size]
                  mov     edi,this_codec_block_size
                  shl     edi,2
                  add     edi,ebx

      @@10:         mov     eax,ebx
                    add     eax,dword ptr rotating_blocks_ptr[2*pointer_size]
                    mov     eax,dword ptr [eax]
                    mov     this_integer,eax

                    fild    this_integer
                    fmul    st,st(3)
                    frndint
                    fmul    st,st(4)

                    fcom    st(2)
                    fnstsw  ax
                    and     ah,$41
                    jnz     @@22

                    fstp    st
                    fld     st(1)
                    add     ecx,1
                    jmp     @@23

      @@22:         fcom    st(1)
                    fnstsw  ax
                    xor     ah,$01
                    and     ah,$41
                    jnz     @@23

                    fstp    st
                    fld     st
                    add     ecx,1

      @@23:         //store bit_removed sample value and calculate
                    //and store sample correction value
                    fist    dword ptr btrd_codec_block[ebx]
                    fmul    scaling_factor_inv
                    frndint
                    fistp   btrd_integer

                    mov     eax,this_integer
                    sub     eax,btrd_integer
                    mov     dword ptr corr_codec_block[ebx],eax

                    add     ebx,integer_size
                    cmp     ebx,edi
                    jl      @@10

                  mov     eax,ecx
                  xor     ecx,ecx
                  cmp     eax,maximum_clips_per_channel
                  jle     @@24

                  fstp    st
                  fstp    st
                  fstp    st
                  fstp    st
                  sub     min_fft_result.btr,1
                  jmp     @@08

      @@24:     add     this_codec_block_clips,eax
                jmp     @@09

      @@25:     fstp    st
                fstp    st
                fstp    st
                fstp    st

                mov     eax,this_codec_block_clips
                add     clipped_samples,eax
                mov     ecx,old_btr
                mov     eax,min_fft_result.btr
                add     total_bits_removed,eax
                sub     ecx,eax
                add     total_bits_lost,ecx
                jmp     @@29

      @@26:     //store codec_block if bits_to_remove is zero or reduced to zero
                //by clipping_prevention algorithm
                mov     ecx,old_btr
                add     total_bits_lost,ecx
                mov     min_fft_result.btr,0

                mov     esi,channels
                xor     edi,edi

      @@27:     sub     esi,1
                jl      @@29
                mov     ebx,dword ptr channel_offset[esi*integer_size]
                mov     ecx,codec_block_size

      @@28:     sub     ecx,1
                jl      @@27

                mov     eax,ebx
                add     eax,dword ptr rotating_blocks_ptr[2*pointer_size]
                mov     eax,dword ptr [eax]
                mov     this_integer,eax
                fild    this_integer
                fmul    scaling_factor
                frndint
                fist    dword ptr btrd_codec_block[ebx]
                fmul    scaling_factor_inv
                frndint
                fistp   btrd_integer

                mov     eax,this_integer
                sub     eax,btrd_integer
                mov     dword ptr corr_codec_block[ebx],eax

                add     ebx,integer_size
                jmp     @@28

      @@29:     pop     edi
                pop     esi
                pop     ebp
                pop     ebx
  end;

  procedure process_this_codec_block;
  var
    analysis_block_number : Integer;
    neg_codec_block_start : Integer;
    pos_codec_block_start : Integer;
  begin
    bits_used_this_codec_block:=0;
    bits_used_this_codec_block_min:=99;
    bits_used_this_codec_block_avg:=0;
    for channel:=0 to channels-1 do
    begin
      bits_used_this_codec_block_channel[channel]:=log2(max(this_codec_block_channel_rms[channel],1));
      bits_used_this_codec_block_avg:=bits_used_this_codec_block_avg+bits_used_this_codec_block_channel[channel]/channels;
      bits_used_this_codec_block_min:=min(bits_used_this_codec_block_min,bits_used_this_codec_block_channel[channel]);
    end;
    bits_used_this_codec_block:=bits_used_this_codec_block_min+(bits_used_this_codec_block_avg-bits_used_this_codec_block_min)*0;

    this_fft_result.btr:=max(0,floor(bits_used_this_codec_block-this_minimum_bits_to_keep));
    min_fft_result.btr:=this_fft_result.btr;

    if bits_used_this_codec_block=0 then
    begin
      for analysis_number:=1 to precalc_analyses do
        for channel:=0 to channels-1 do
        begin
          saved_fft_results[channel,analysis_number].btr:=-1;
          saved_fft_results[channel,analysis_number].start:=-9999;
        end;
      empty_codec_blocks:=empty_codec_blocks+1;
      min_fft_result.btr:=0;
    end
    else
    begin
      for analysis_number:=1 to precalc_analyses do
        if (this_fft_string[analysis_number]='1') and (fft_length[analysis_number] div 2<=codec_block_size)then
        begin
          this_fft_length:=fft_length[analysis_number];
          this_fft_length_recip:=fft_length_recip[analysis_number];

          this_fft_result.analysis:=analysis_number;
          this_fft_result.nminbin:=-1;

          if SetFFTLength(this_fft_length div 2)=0 then
            lossyWAVerror('Fatal FFT error!',$88);

          neg_codec_block_start:=-(last_codec_block_size+prev_codec_block_size);
          pos_codec_block_start:=this_codec_block_size+next_codec_block_size-this_fft_length;

          this_actual_analysis_blocks_start:=actual_analysis_blocks_start[analysis_number];
          this_fft_underlap_length:=fft_underlap_length[analysis_number];

          for analysis_block_number:=0 to analysis_blocks[analysis_number] do
          begin
            actual_analysis_block_start:=floor(this_actual_analysis_blocks_start+analysis_block_number*this_fft_underlap_length);
            actual_analysis_block_start:=min(max(actual_analysis_block_start,neg_codec_block_start),pos_codec_block_start);
            for channel:=0 to channels-1 do
            begin
              if (analysis_block_number=0) then
              begin
                if (actual_analysis_block_start=saved_fft_results[channel,analysis_number].start) and (saved_fft_results[channel,analysis_number].btr<>-1) then
                  min_fft_result.btr:=min(min_fft_result.btr,saved_fft_results[channel,analysis_number].btr)
              end
              else
              begin
                Fill_FFT_Input;
                FFT_Real;
                this_fft_result.spread:=spread_complex;
                this_fft_result.btr:=threshold_index[analysis_number,this_fft_result.spread];

                if this_fft_result.btr<min_fft_result.btr then
                  min_fft_result:=this_fft_result;

                if analysis_block_number=analysis_blocks[analysis_number] then
                begin
                  this_fft_result.start:=actual_analysis_block_start-codec_block_size;
                  saved_fft_results[channel,analysis_number]:=this_fft_result;
                end;
              end;
            end;
          end;

          if min_fft_result.sminbin>min_fft_result.savebin then
            fft_average:=fft_average+1
          else
            fft_minimum:=fft_minimum+1;
        end;
      end;

    remove_bits;

    if (detailled_output=1) and (silent=0) then
      remove_bits_output
    else
      inc(current_codec_block);

    if writeNextBtrdCodecBlock=0 then
      lossyWAVerror('Error writing to output file.',$21);

    if correct=1 then
      if writeNextCorrCodecBlock=0 then
        lossyWAVerror('Error writing to correction file.',$22);

    samples_processed:=(blocks_read-1)*codec_block_size+this_codec_block_size;

    if (detailled_output=0) and (verbosity=1) and (silent=0) and ((last_print=0) or (blocks_read mod half_mb_blocks=0)) then
    begin
      Get_Elapsed_Time;
      write(#13,'Progress   : ',samples_processed/totalsamples*100:0:2,'%, ',
                                powersoftwoin[20]*samples_processed*channels*bytespersample:0:2,'MB, ',
                                total_bits_removed/blocks_read:0:4,' bits; ',
                                processing_rate:0:2,'x;');
      if (stimer.elapsed>2) and (samples_processed>0) then
        write(' ',stimer.elapsed:0:2,'s/',totalsamples/samples_processed*stimer.elapsed:0:2,'s;    ');
      last_print:=blocks_read;
    end;
  end;

  procedure shift_codec_blocks;
  asm
                mov     eax,dword ptr rotating_blocks_ptr[3*pointer_size]
                xchg    eax,dword ptr rotating_blocks_ptr[2*pointer_size]
                xchg    eax,dword ptr rotating_blocks_ptr[1*pointer_size]
                xchg    eax,dword ptr rotating_blocks_ptr[0*pointer_size]
                mov     dword ptr rotating_blocks_ptr[3*pointer_size],eax

                xor     eax,eax
                xchg    eax,next_codec_block_size
                xchg    eax,this_codec_block_size
                xchg    eax,last_codec_block_size
                mov     prev_codec_block_size,eax

                push   ebx
                push   ebp
                push   esi
                push   edi

                mov    ebp,dword ptr rotating_blocks_ptr[2*pointer_size]
                mov    eax,this_codec_block_size
                cmp    eax,1
                jl     @@02

                fld1
                fild   this_codec_block_size
                fdiv
                fstp   this_codec_block_size_recip

                add    blocks_read,1
                mov    ebx,channels

      @@00:     sub    ebx,1
                jl     @@02

                  fldz
                  mov    ecx,dword ptr channel_offset[ebx*integer_size]
                  add    ecx,ebp
                  mov    edx,eax

      @@01:       sub    edx,1
                  jl     @@01a

                  fild   dword ptr [ecx+edx*integer_size]
                  fmul   st,st
                  fadd

                  jmp    @@01

      @@01a:    fmul   this_codec_block_size_recip
                fsqrt
                fstp   qword ptr this_codec_block_channel_rms[ebx*real_size]

                jmp    @@00

      @@02:     pop    edi
                pop    esi
                pop    ebp
                pop    ebx
  end;

//==============================================================================
begin
//==============================================================================
  check_switches;

  if merging=1 then
  begin
    mergefiles;
  end
  else
  begin
    if openWavIO(WAVInp, lossyOut, codec_block_size)=0 then
      lossyWAVerror('Error initialising wavIO unit.',$11);

    initial_setup;

    if readNextNextCodecBlock=0 then
      lossyWAVerror('Error reading from input file.',$21);
    shift_codec_blocks;
    dec(blocks_read);

    while this_codec_block_size>0 do
    begin
      success := readNextNextCodecBlock;
      next_codec_block_size:=next_codec_block_size*success;
      if this_codec_block_size>0 then
        process_this_codec_block;
      shift_codec_blocks;
    end;

    if closeWavIO=0 then
      lossyWAVerror('Error closing wavIO unit.',$11);
    if (silent=0) then
      write_cleanup;
  end;
//==============================================================================
end.
//==============================================================================

