//==============================================================================
//
//    lossyWAV: Added noise WAV bit reduction method by David Robinson;
//    Copyright (C) 2007, 2008 Nick Currie, Copyleft.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//    Contact: lossywav <at> yahoo <dot> com
//
//==============================================================================
unit nrand;
//==============================================================================
{$DEFINE CPU386}
{$N+ $A8 $I-}

//==============================================================================
interface
//==============================================================================
uses
  Windows,
  ntypes;

  //============================================================================
  // Fibonacci Shift Register pseudo random number generator.
  //============================================================================
  function  fib_rand_int    : Integer;  //32-bit fibonacci shift random integer;
  function  fib_rand_float  : extended; //32-bit fibonacci shift random float;
  function  tpdf_rand_float : Extended; //32-bit fibonacci TPDF random float;

//==============================================================================
implementation
//==============================================================================

//==============================================================================
//nrand constants, variables & types
//==============================================================================
const
  tpdf_rand_and_val     = $7FFF0000;
  tpdf_rand_shift       = 16;
Var
  QPC_Start : Int64;

  fib_rand_int_value      : Integer = 0;
  last_fib_rand_int_value : Integer = 0;
  tpdf_rand_int_value     : Integer = 0;

  fib_rand_float_value    : Double = 0;
  tpdf_rand_float_value   : Double = 0;
  ex_tpdf_rand_divisor    : double = 1/tpdf_rand_and_val;

  nrand_packer            : array[0..1] of integer;

  function fib_rand_int : Integer; //32-bit resolution fibonacci shift random integer;
  asm
                push    ecx
                push    edx

                mov     eax,fib_rand_int_value
                mov     last_fib_rand_int_value,eax
                mov     ecx,eax
                mov     edx,eax

                shr     ecx,6
                xor     edx,ecx
                shr     ecx,1
                xor     edx,ecx
                shr     ecx,2
                xor     edx,ecx
                and     edx,$003FFFFF
                shl     eax,22
                add     eax,edx
                mov     fib_rand_int_value,eax

                pop     edx
                pop     ecx
  end;

  function fib_rand_float : extended; //32-bit resolution fibonacci shift random float;
  asm
                push    eax
                call    fib_rand_int
                fild    fib_rand_int_value
                fmul    qword ptr powersoftwoin[32*real_size]
                fadd    qword ptr powersoftwoin[1*real_size]
                fst     fib_rand_float_value
                pop     eax
  end;

  function tpdf_rand_float : Extended; //32-bit resolution TPDF random float;
  asm
                push    eax
                call    fib_rand_int
                fild    fib_rand_int_value
                fild    last_fib_rand_int_value
                fsub
                fmul    qword ptr powersoftwoin[32*real_size]
                fst     tpdf_rand_float_value
                pop     eax
  end;

//==============================================================================
// Unit nrand end.
//==============================================================================
begin
  QueryPerformanceCounter(QPC_Start);
  fib_rand_int_value:=((QPC_Start shr 16) and $FFFFFFFF);
  if fib_rand_int_value=0 then
    fib_rand_int_value:=$4F1BBCDD; // Golden Ratio x 2^31;
  nrand_packer[0]:=0;
end.
//==============================================================================

